from __future__ import annotations


from dataclasses import dataclass
from enum import Enum
from typing import Iterable
import subprocess
import neaty.log


def _noop(*args, **kwargs) -> None:
    pass


def _join_lines(lines: Iterable[str]) -> str:
    return ''.join([ln + '\n' for ln in lines])


@dataclass(frozen=True, slots=True)
class Result:
    action: Action
    out: str
    err: str
    es: int

    def is_failure(self) -> bool:
        return self.es > 0

    def is_success(self) -> bool:
        return self.es == 0

    def _fmt_payload(self) -> str:
        verb = "succeeded"
        if self.is_failure():
            verb = "exited with: %d" % self.es
        if self.action.action_type == ActionType.COMMAND:
            return ("%s: command finished: %s (unquoted here); %s"
                    % (self.action.name, verb, " ".join(self.action.args)))
        if self.action.action_type == ActionType.SCRIPT:
            return "%s: script finished: %s" % (self.action.name, verb)
        return "%s: finished: %s" % (self.action.name, verb)

    def _log_failing(self) -> None:
        neaty.log.warn(self._fmt_payload())
        for line in self.out.splitlines():
            neaty.log.warn("  out:|" + line)
        for line in self.err.splitlines():
            neaty.log.warn("  err:|" + line)

    def _log_successful(self) -> None:
        neaty.log.think(self._fmt_payload())
        for line in self.err.splitlines():
            neaty.log.warn(line)

    def log(self) -> None:
        fn = self._log_successful if self.es == 0 else self._log_failing
        fn()


class ActionType(Enum):
    COMMAND = 'command'
    SCRIPT = 'script'


@dataclass(frozen=True, slots=True)
class Action:
    args: list[str]
    name: str
    dry: bool
    action_type: ActionType

    @classmethod
    def from_code(cls,
                  code: str,
                  name: str,
                  dry: bool = False,
                  ):
        return cls(
            args=['sh', '-c', code],
            name=name,
            dry=dry,
            action_type=ActionType.SCRIPT,
        )

    @classmethod
    def from_cmd(cls,
                 cmd: list[str],
                 name: str,
                 dry: bool = False,
                 ):
        return cls(
            args=cmd,
            name=name,
            dry=dry,
            action_type=ActionType.COMMAND,
        )

    def run(self) -> Result:
        if self.dry:
            return Result(
                action=self,
                out='',
                err='',
                es=0,
            )
        p = subprocess.Popen(
            self.args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = p.communicate()
        return Result(
            action=self,
            out=out.decode(),
            err=err.decode(),
            es=p.returncode,
        )


def _describe_cmd(cmd: list[str],
                  name: str,
                  ) -> Iterable[str]:
    togo = cmd.copy()
    yield "-- begin cmd %s --" % name
    if len(togo) == 0:
        yield '# no command passed to run_cmd()'
    elif len(togo) == 1:
        yield quote(togo[0])
    else:
        yield togo.pop(0) + '\\'
        while len(togo) > 1:
            yield '  ' + togo.pop(0) + '\\'
        yield '  ' + togo.pop(0)
    yield "-- end cmd %s --" % name


def _describe_code(code: str,
                   name: str,
                   ) -> Iterable[str]:
    yield "-- begin code %s --" % name
    for line in code.split('\n'):
        yield '|' + line.rstrip('\n')
    yield "-- end code %s --" % name


def stdout_of(cmd: list[str],
              name: str = 'stdout_of()',
              ) -> str:
    result = Action.from_cmd(cmd, name).run()
    if not result.is_success():
        result.log()
        raise RuntimeError("command failed")
    return result.out


def stdout_of_code(code: str,
                   name: str = 'stdout_of_code()',
                   ) -> str:
    result = Action.from_code(code, name).run()
    if not result.is_success():
        result.log()
        raise RuntimeError("command failed")
    return result.out


def run_cmd(cmd: list[str],
            name: str = 'run_cmd()',
            dry: bool = False,
            ) -> bool:
    if dry:
        for line in _describe_cmd(cmd, name):
            neaty.log.warn(line)
        return True
    result = Action.from_cmd(cmd, name or 'run_cmd()').run()
    result.log()
    return result.is_success()


def run_code(code: str,
             name: str = 'run_code()',
             dry: bool = False,
             ) -> bool:
    if dry:
        for line in _describe_code(code, name):
            neaty.log.warn(line)
        return True
    result = Action.from_code(code, name).run()
    result.log()
    return result.is_success()


def dry_run_code(code: str,
                 name: str = 'dry_run_code()',
                 ):
    for line in _describe_code(code, name):
        neaty.log.warn(line)


def quote_words(words: Iterable[str]) -> list[str]:
    for word in words:
        if '\n' in word:
            raise ValueError(f"newlines are not allowed: word={word!r}")
    cmd = ['bash', '-c', "printf '%q\\n' \"$@\"", '--']
    cmd.extend(words)
    return stdout_of(cmd).splitlines()


def quote(text: str) -> str:
    return stdout_of([
        'bash', '-c', 'printf %q "$1"', '--', text
    ])


def content_equals(path_a: str,
                   path_b: str,
                   ) -> bool:
    with open(path_a) as f:
        ca = f.read()
    with open(path_b) as f:
        cb = f.read()
    return ca == cb
